from pydantic import BaseModel, Field
from typing import Dict, List, Union

from fastapi import FastAPI, HTTPException
from .accessors import generate_file_accessor
from .models import BinaryThresholdSegmentationModel
from .pipelines.shared import PipelineRecord
from .roiset import IntensityThresholdInstanceMaskSegmentationModel, RoiSetExportParams, SerializeRoiSetError
from .session import session, AccessorIdError, InvalidPathError, RoiSetIdError, RunTaskError, WriteAccessorError

app = FastAPI(debug=True)

from .pipelines.router import router
app.include_router(router)


@app.on_event("startup")
def startup():
    pass


@app.get('/')
def read_root():
    return {'success': True}


@app.get('/paths/session')
def get_top_session_path():
    return session.get_paths()['session']


@app.get('/paths/inbound')
def get_inbound_path():
    return session.get_paths()['inbound_images']


@app.get('/paths/outbound')
def get_outbound_path():
    return session.get_paths()['outbound_images']


@app.get('/paths')
def list_session_paths():
    return session.get_paths()


@app.get('/status')
def show_session_status():
    return {
        'status': 'running',
        'models': session.describe_loaded_models(),
        'paths': session.get_paths(),
        'rois': session.list_rois(),
        'accessors': session.list_accessors(),
        'tasks': session.tasks.list_tasks(),
    }


def _change_path(key, path):
    try:
        session.set_data_directory(key, path)
    except InvalidPathError as e:
        raise HTTPException(404, f'Did not find valid folder at: {path}')


@app.put('/paths/watch_input')
def watch_input_path(path: str):
    return _change_path('inbound_images', path)


@app.put('/paths/watch_output')
def watch_output_path(path: str):
    return _change_path('outbound_images', path)


@app.get('/session/restart')
def restart_session() -> dict:
    session.restart()
    return session.describe_loaded_models()


@app.get('/session/logs')
def list_session_log() -> list:
    return session.get_log_data()


@app.get('/models')
def list_active_models():
    return session.describe_loaded_models()


class BinaryThresholdSegmentationParams(BaseModel):
    tr: Union[int, float] = Field(0.5, description='Threshold for binary segmentation')


@app.put('/models/seg/threshold/load/')
def load_binary_threshold_model(p: BinaryThresholdSegmentationParams, model_id=None) -> dict:
    result = session.load_model(BinaryThresholdSegmentationModel, key=model_id, params=p)
    session.log_info(f'Loaded binary threshold segmentation model {result}')
    return {'model_id': result}


@app.put('/models/classify/threshold/load')
def load_intensity_threshold_instance_segmentation_model(p: BinaryThresholdSegmentationParams, model_id=None) -> dict:
    result = session.load_model(IntensityThresholdInstanceMaskSegmentationModel, key=model_id, params=p)
    session.log_info(f'Loaded permissive instance segmentation model {result}')
    return {'model_id': result}


@app.get('/accessors')
def list_accessors():
    return session.list_accessors()


def _session_accessor(func, acc_id):
    try:
        return func(acc_id)
    except AccessorIdError as e:
        raise HTTPException(404, f'Did not find accessor with ID {acc_id}')


@app.get('/accessors/{accessor_id}')
def get_accessor(accessor_id: str):
    return _session_accessor(session.get_accessor_info, accessor_id)


@app.get('/accessors/delete/{accessor_id}')
def delete_accessor(accessor_id: str):
    if accessor_id == '*':
        return session.del_all_accessors()
    else:
        return _session_accessor(session.del_accessor, accessor_id)

# TODO: optional lazy loading, so that batch task can be queued before file data is loaded
@app.put('/accessors/read_from_file/{filename}')
def read_accessor_from_file(filename: str, accessor_id: Union[str, None] = None):
    fp = session.paths['inbound_images'] / filename
    if not fp.exists():
        raise HTTPException(status_code=404, detail=f'Could not find file:\n{filename}')
    acc = generate_file_accessor(fp)
    return session.add_accessor(acc, accessor_id=accessor_id)


@app.put('/accessors/write_to_file/{accessor_id}')
def write_accessor_to_file(accessor_id: str, filename: Union[str, None] = None, pop: bool = True) -> str:
    try:
        return session.write_accessor(accessor_id, filename, pop=pop)
    except AccessorIdError as e:
        raise HTTPException(404, f'Did not find accessor with ID {accessor_id}')
    except WriteAccessorError as e:
        raise HTTPException(409, str(e))

@app.get('/rois')
def list_rois():
    return session.list_rois()

def _session_roiset(func, roiset_id):
    try:
        return func(roiset_id)
    except RoiSetIdError as e:
        raise HTTPException(404, f'Did not find RoiSet with ID {roiset_id}')

@app.get('/rois/{roiset_id}')
def get_roiset(roiset_id: str):
    return _session_roiset(session.get_roiset_info, roiset_id)

class RoiSetSerializationRecord(BaseModel):
    dataframe: str
    tight_patch_masks: Union[List[str], None]

@app.put('/rois/write/{roiset_id}')
def write_roiset_to_file(roiset_id: str, where: Union[str, None] = None) -> RoiSetSerializationRecord:
    try:
        return session.write_roiset(roiset_id, where)
    except RoiSetIdError as e:
        raise HTTPException(404, f'Did not find RoiSet with ID {roiset_id}')
    except SerializeRoiSetError as e:
        raise HTTPException(409, str(e))

@app.put('/rois/products/{roiset_id}')
def roiset_export_products(
        roiset_id: str,
        channel: Union[int, None] = None,
        params: RoiSetExportParams = RoiSetExportParams()
) -> dict:
    roiset = _session_roiset(session.get_roiset, roiset_id)
    products = roiset.get_export_product_accessors(channel, params)
    return {k: session.add_accessor(v, f'{roiset_id}_{k}') for k, v in products.items()}

@app.put('/rois/obmap/{roiset_id}/{model_id}')
def roiset_get_object_map(
        roiset_id: str,
        model_id: str,
        channel: Union[int, None] = None,
) -> str:
    products = roiset_export_products(
        roiset_id,
        channel,
        RoiSetExportParams(object_classes=True),
    )
    for k, rid in products.items():
        if k == f'object_classes_{model_id}':
            return rid
    raise HTTPException(
        404,
        f'Did not find object map from classification model {model_id} in RoiSet {roiset_id}'
    )

class TaskInfo(BaseModel):
    module: str
    params: dict
    func_str: str
    status: str
    error: Union[str, None]
    result: Union[Dict, None]

@app.put('/tasks/{task_id}/run')
def run_task(task_id: str) -> PipelineRecord:
    try:
        return session.tasks.run_task(task_id)
    except RunTaskError as e:
        raise HTTPException(409, str(e))

@app.get('/tasks/{task_id}')
def get_task(task_id: str) -> TaskInfo:
    return session.tasks.get_task_info(task_id)

@app.get('/tasks')
def list_tasks() -> Dict[str, TaskInfo]:
    res = session.tasks.list_tasks()
    return res

# TODO: implement
@app.put('/tasks/run/on_files')
def task_file_batch():
    # new callable that parameterizes file name to acc_id, then passes this to to /tasks/*/run
    pass